@@ -1,5 +1,5 @@
 /************************************************************************************
-   Copyright (C) 2013, 2023 MariaDB Corporation AB
+   Copyright (C) 2013, 2025 MariaDB Corporation plc
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -24,23 +24,6 @@
 
 #define MADB_FIELD_IS_BINARY(_field) ((_field)->charsetnr == BINARY_CHARSETNR)
 
-void MADB_NewStmtHandle(MADB_Stmt *Stmt)
-{
-  // static const my_bool UpdateMaxLength= 1;
-  Stmt->stmt.reset();
-
-  /*if (stmt != NULL)
-  {
-    mysql_stmt_attr_set(stmt, STMT_ATTR_UPDATE_MAX_LENGTH, &UpdateMaxLength);
-  }
-  else
-  {
-    MADB_SetError(&Stmt->Error, MADB_ERR_HY001, NULL, 0);
-  }
-
-  return stmt;*/
-}
-
 /* Required, but not sufficient condition */
 bool QueryIsPossiblyMultistmt(MADB_QUERY *Query)
 {
@@ -111,7 +94,7 @@ int MADB_KeyTypeCount(MADB_Dbc *Connection, char *TableName, int *PrimaryKeysCou
   p+= _snprintf(p, sizeof(StmtStr) - (p - StmtStr), "%s LIMIT 0", TableName);
   std::lock_guard<std::mutex> localScopeLock(Connection->guard->getLock());
   Connection->guard->safeRealQuery({StmtStr, (std::size_t)(p - StmtStr)});
-  if ((Res= mysql_store_result(Connection->mariadb)) != NULL)
+  if ((Res= mysql_store_result(Connection->mariadb)) != nullptr)
   {
     Count= mysql_field_count(Connection->mariadb);
     for (i= 0; i < Count; ++i)
@@ -263,7 +246,7 @@ const char *MADB_GetTypeName(const MYSQL_FIELD *Field)
 }
 /* }}} */
 
-
+/* {{{ MADB_GetDefaultColumnValues */
 MYSQL_RES *MADB_GetDefaultColumnValues(MADB_Stmt *Stmt, const MYSQL_FIELD *fields)
 {
   SQLString DynStr("SELECT COLUMN_NAME, COLUMN_DEFAULT FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA='");
@@ -272,9 +255,9 @@ MYSQL_RES *MADB_GetDefaultColumnValues(MADB_Stmt *Stmt, const MYSQL_FIELD *field
   DynStr.reserve(512);
   DynStr.append(fields[0].db).append("' AND TABLE_NAME='").append(fields[0].org_table).append("' AND COLUMN_NAME IN (");
 
-  for (i = 0; i < Stmt->metadata->getColumnCount(); i++)
+  for (i= 0; i < Stmt->metadata->getColumnCount(); i++)
   {
-    MADB_DescRecord* Rec = MADB_DescGetInternalRecord(Stmt->Ard, i, MADB_DESC_READ);
+    MADB_DescRecord* Rec= MADB_DescGetInternalRecord(Stmt->Ard, i, MADB_DESC_READ);
 
     if (!Rec->inUse || MADB_ColumnIgnoredInAllRows(Stmt->Ard, Rec) == TRUE)
     {
@@ -288,7 +271,7 @@ MYSQL_RES *MADB_GetDefaultColumnValues(MADB_Stmt *Stmt, const MYSQL_FIELD *field
   Stmt->Connection->guard->safeRealQuery(DynStr);
   return mysql_store_result(Stmt->Connection->mariadb);
 }
-
+/* }}} */
 
 char *MADB_GetDefaultColumnValue(MYSQL_RES *res, const char *Column)
 {
@@ -480,7 +463,8 @@ size_t MADB_GetOctetLength(const MYSQL_FIELD *Field, unsigned short MaxCharLen)
 }
 /* }}} */
 
-/* {{{ MADB_GetDefaultType */
+/* {{{ MADB_GetDefaultType
+ */
 int MADB_GetDefaultType(int SQLDataType)
 {
   switch(SQLDataType)
@@ -536,9 +520,10 @@ int MADB_GetDefaultType(int SQLDataType)
 }
 /* }}} */
 
-/* {{{ MapMariadDbToOdbcType */
-       /* It's not quite right to mix here C and SQL types, even though constants are sort of equal */
-SQLSMALLINT MapMariadDbToOdbcType(const MYSQL_FIELD *field)
+/* {{{ MapMariadDbToOdbcType
+ * @param Stmt - we need it for the information that may need to do some ajustments in the mappings 
+ */
+SQLSMALLINT MapMariadDbToOdbcType(const MADB_Stmt* Stmt, const MYSQL_FIELD *field)
 {
   switch (field->type) {
     case MYSQL_TYPE_BIT:
@@ -573,7 +558,15 @@ SQLSMALLINT MapMariadDbToOdbcType(const MYSQL_FIELD *field)
     case MYSQL_TYPE_LONG_BLOB:
       return MADB_FIELD_IS_BINARY(field) ? SQL_LONGVARBINARY : SQL_LONGVARCHAR;
     case MYSQL_TYPE_LONGLONG:
-      return SQL_BIGINT;
+      /* Treating MSAccess differently */
+      if (Stmt->Connection->Dsn->NoBigint)
+      {
+        return SQL_INTEGER;
+      }
+      else
+      {
+        return Stmt->Connection->Environment->AppType == ATypeMSAccess ? SQL_VARCHAR : SQL_BIGINT;
+      }
     case MYSQL_TYPE_STRING:
       return MADB_FIELD_IS_BINARY(field) ? SQL_BINARY : SQL_CHAR;
     case MYSQL_TYPE_VAR_STRING:
@@ -753,12 +746,18 @@ SQLRETURN MADB_CopyMadbTimestamp(MADB_Stmt *Stmt, MYSQL_TIME *tm, SQLPOINTER Dat
           ts->month= cur_tm->tm_mon + 1;
           ts->day= cur_tm->tm_mday;
           ts->fraction= 0;
+          /* If we allowing fraction in time on the way to the server, we should allow it on the way back
+           */
+          if (Stmt->Connection->Dsn->AllowDtTruncation) {
+            ts->fraction= tm->second_part * 1000;
+          }
         }
         else
         {
           ts->year= tm->year;
           ts->month= tm->month;
           ts->day= tm->day;
+          // Microseconds of MYSQL_TIME to nanoseconds of SQL_TIMESTAMP_STRUCT
           ts->fraction= tm->second_part * 1000;
         }
         ts->hour= tm->hour;
@@ -870,14 +869,14 @@ void* GetBindOffset(MADB_Header& Header, void* Ptr, SQLULEN RowNumber, std::size
   if (Header.BindType == SQL_BIND_BY_COLUMN ||
     Header.BindType == SQL_PARAM_BIND_BY_COLUMN)
   {
-    BindOffset+= PtrSize * RowNumber;
+    BindOffset+= PtrSize*RowNumber;
   }
   else
   {
-    BindOffset+= Header.BindType * RowNumber;
+    BindOffset+= Header.BindType*RowNumber;
   }
 
-  return (char *)Ptr + BindOffset;
+  return (char*)Ptr + BindOffset;
 }
 
 /* Checking if column ignored in all bound rows. Should hel*/
@@ -926,7 +925,7 @@ int MADB_CharToSQLNumeric(char *buffer, MADB_Desc *Ard, MADB_DescRecord *ArdReco
   MADB_NumericInit(number, ArdRecord);
 
   /* Determining the sign of the number. From now on we dean with unsigned number */
-  if (!(number->sign = (*p == '-') ? 0 : 1))
+  if (!(number->sign= (*p == '-') ? 0 : 1))
   {
     p++;
   }
@@ -978,7 +977,7 @@ int MADB_CharToSQLNumeric(char *buffer, MADB_Desc *Ard, MADB_DescRecord *ArdReco
       if (number->scale < 0 && (number->precision - number->scale) >= digits_count)
       {
         /* Checking that all digits past presision are '0'. Otherwise - overflow */
-        for (i = digits_count - number->precision; i > 0; --i)
+        for (i= digits_count - number->precision; i > 0; --i)
         {
           if (*(p + digits_count - i) != '0')
           {
@@ -1068,13 +1067,13 @@ int MADB_CharToSQLNumeric(char *buffer, MADB_Desc *Ard, MADB_DescRecord *ArdReco
     }
 
     digits_count= MIN(digits_count, MADB_DEFAULT_PRECISION + 1);
-    for (hval = 0, bit = 1L, sta = 0, olen = 0; sta < digits_count;)
+    for (hval= 0, bit= 1L, sta= 0, olen= 0; sta < digits_count;)
     {
-      for (dig = 0, i = sta; i < digits_count; i++)
+      for (dig= 0, i= sta; i < digits_count; i++)
       {
-        tv = dig * 10 + digits[i] - '0';
-        dig = tv % 2;
-        digits[i] = tv / 2 + '0';
+        tv= dig * 10 + digits[i] - '0';
+        dig= tv % 2;
+        digits[i]= tv / 2 + '0';
         if (i == sta && tv < 2)
           sta++;
       }
@@ -1085,13 +1084,13 @@ int MADB_CharToSQLNumeric(char *buffer, MADB_Desc *Ard, MADB_DescRecord *ArdReco
       {
         if (olen >= SQL_MAX_NUMERIC_LEN)
         {
-          //number->scale = sta - number->precision;
+          //number->scale= sta - number->precision;
           ret= MADB_ERR_22003;
           break;
         }
-        number->val[olen++] = hval;
-        hval = 0;
-        bit = 1L;
+        number->val[olen++]= hval;
+        hval= 0;
+        bit= 1L;
 
       } 
     }
@@ -1099,7 +1098,7 @@ int MADB_CharToSQLNumeric(char *buffer, MADB_Desc *Ard, MADB_DescRecord *ArdReco
     {
       if (olen < SQL_MAX_NUMERIC_LEN)
       {
-        number->val[olen++] = hval;
+        number->val[olen++]= hval;
       }
       else
       {
@@ -1142,10 +1141,11 @@ SQLRETURN MADB_DaeStmt(MADB_Stmt *Stmt, SQLUSMALLINT Operation)
   MADB_CLEAR_ERROR(&Stmt->Error);
 
   if (Stmt->DaeStmt)
-    Stmt->Methods->StmtFree(Stmt->DaeStmt, SQL_DROP);
-  Stmt->DaeStmt= nullptr;
+  {
+    MADB_DeleteDaeStmt(Stmt);
+  }
 
-  if (!SQL_SUCCEEDED(MA_SQLAllocHandle(SQL_HANDLE_STMT, (SQLHANDLE)Stmt->Connection, (SQLHANDLE *)&Stmt->DaeStmt)))
+  if (!SQL_SUCCEEDED(MADB_StmtInit(Stmt->Connection, (SQLHANDLE *)&Stmt->DaeStmt, false)))
   {
     return MADB_CopyError(&Stmt->Error, &Stmt->Connection->Error);
   }
@@ -1184,7 +1184,7 @@ SQLRETURN MADB_DaeStmt(MADB_Stmt *Stmt, SQLUSMALLINT Operation)
   if (!SQL_SUCCEEDED(Stmt->DaeStmt->Prepare(Query.c_str(), (SQLINTEGER)Query.length(), true)))
   {
     MADB_CopyError(&Stmt->Error, &Stmt->DaeStmt->Error);
-    Stmt->Methods->StmtFree(Stmt->DaeStmt, SQL_DROP);
+    MADB_DeleteDaeStmt(Stmt);
   }
   return Stmt->Error.ReturnValue;
 }
@@ -1202,7 +1202,7 @@ int MADB_FindNextDaeParam(MADB_Desc *Desc, int InitialParam, SQLSMALLINT RowNumb
       if (Record->OctetLengthPtr)
       {
         /* Stmt->DaeRowNumber is 1 based */
-        SQLLEN *OctetLength = (SQLLEN *)GetBindOffset(Desc->Header, Record->OctetLengthPtr, RowNumber > 1 ? RowNumber - 1 : 0, sizeof(SQLLEN));
+        SQLLEN *OctetLength= (SQLLEN *)GetBindOffset(Desc->Header, Record->OctetLengthPtr, RowNumber > 1 ? RowNumber - 1 : 0, sizeof(SQLLEN));
         if (PARAM_IS_DAE(OctetLength))
         {
           return i;
@@ -1214,22 +1214,27 @@ int MADB_FindNextDaeParam(MADB_Desc *Desc, int InitialParam, SQLSMALLINT RowNumb
   return MADB_NOPARAM;
 }
 
-
-BOOL MADB_IsNumericType(SQLSMALLINT ConciseType)
+/* Returns true if the type is */
+bool MADB_IsDecimalType(SQLSMALLINT ConciseType)
 {
   switch (ConciseType)
   {
     case SQL_C_DOUBLE:
     case SQL_C_FLOAT:
     case SQL_DECIMAL:
+    case SQL_C_NUMERIC:
       return TRUE;
   }
+  return FALSE;
+}
 
-  return MADB_IsIntType(ConciseType);
+bool MADB_IsNumericType(SQLSMALLINT ConciseType)
+{
+  return MADB_IsDecimalType(ConciseType) || MADB_IsIntType(ConciseType);
 }
 
 
-BOOL MADB_IsIntType(SQLSMALLINT ConciseType)
+bool MADB_IsIntType(SQLSMALLINT ConciseType)
 {
   switch (ConciseType)
   {
@@ -1257,3 +1262,27 @@ enum enum_field_types MADB_GetNativeFieldType(MADB_Stmt *Stmt, uint32_t i)
   return static_cast<enum enum_field_types>(Stmt->metadata->getColumnType(i));
 }
 
+
+
+SQLRETURN MADB_KillAtServer(MADB_Dbc *Connection, MADB_Error* Error)
+{
+  MYSQL *MariaDb, *Kill=Connection->mariadb;
+
+  if (!(MariaDb= mysql_init(NULL)))
+  {
+    return MADB_SetError(Error, MADB_ERR_HY001, NULL, 0);
+  }
+  if (SQL_SUCCEEDED(Connection->CoreConnect(MariaDb, Connection->Dsn, Error, 0)))
+  {
+    char StmtStr[32];
+    // 32 is enough to fit the query with longest 8byte unsigned  thread id, thus sprintf is fine here
+    int len= sprintf(StmtStr, "KILL QUERY %lu", mysql_thread_id(Kill));
+    if (mysql_real_query(MariaDb, StmtStr, len))
+    {
+      mysql_close(MariaDb);
+      return MADB_SetError(Error, MADB_ERR_HY000, "Error while terminating the process on the server", 0);
+    }
+  }
+  mysql_close(MariaDb);
+  return SQL_SUCCESS;
+}
