@@ -1,5 +1,5 @@
 /************************************************************************************
-   Copyright (C) 2013,2023 MariaDB Corporation AB
+   Copyright (C) 2013,2025 MariaDB Corporation plc
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -23,6 +23,7 @@
 #include <memory>
 #include <list>
 #include <map>
+#include <atomic>
 #include "ma_c_stuff.h"
 #include "ma_legacy_helpers.h"
 #include "lru/pscache.h"
@@ -30,6 +31,7 @@
 #include "template/CArray.h"
 #include "class/pimpls.h"
 #include "class/Results.h"
+#include "mariadb/sqlmariadb.h"
 
 namespace mariadb
 {
@@ -192,12 +194,6 @@ typedef struct
   my_bool AppType;      /* Allocated by Application ? */
 } MADB_Desc;
 
-struct st_ma_desc_fldid
-{
-  SQLSMALLINT FieldIdentifier;
-  SQLSMALLINT Access[4];
-};
-
 struct st_ma_stmt_methods;
 
 typedef struct 
@@ -215,6 +211,8 @@ typedef struct
   SQLUINTEGER RetrieveData;
 	SQLUINTEGER UseBookmarks;
   SQLSMALLINT BookmarkType;
+  bool ExecDirectOnServer;
+  bool PrepareOnClient;
 } MADB_StmtOptions;
 
 /* TODO: To check is it 0 or 1 based? not quite clear from its usage */
@@ -228,8 +226,11 @@ typedef struct
 
 enum MADB_DaeType {MADB_DAE_NORMAL=0, MADB_DAE_ADD=1, MADB_DAE_UPDATE=2, MADB_DAE_DELETE=3};
 
-#define RESET_DAE_STATUS(Stmt_Hndl) (Stmt_Hndl)->Status=0; (Stmt_Hndl)->PutParam= -1
-#define MARK_DAE_DONE(Stmt_Hndl)    (Stmt_Hndl)->Status=0; (Stmt_Hndl)->PutParam= (Stmt_Hndl)->ParamCount
+#define MADB_NO_DATA_NEEDED -2
+#define MADB_NEED_DATA -1
+
+#define RESET_DAE_STATUS(Stmt_Hndl) (Stmt_Hndl)->PutParam= MADB_NO_DATA_NEEDED
+#define MARK_DAE_DONE(Stmt_Hndl)    (Stmt_Hndl)->PutParam= (Stmt_Hndl)->ParamCount
 
 #define PARAM_IS_DAE(Len_Ptr) ((Len_Ptr) && (*(Len_Ptr) == SQL_DATA_AT_EXEC || *(Len_Ptr) <= SQL_LEN_DATA_AT_EXEC_OFFSET))
 #define DAE_DONE(Stmt_Hndl) ((Stmt_Hndl)->PutParam >= (Stmt_Hndl)->ParamCount)
@@ -286,20 +287,22 @@ private:
 // Stmt has to know few things about connection and descriptor
 #include "ma_connection.h"
 #include "ma_desc.h"
-
+#define xxx sizeof(std::mutex)
 struct MADB_Stmt
 {
+  MADB_QUERY                Query;
   MADB_StmtOptions          Options;
   MADB_Error                Error;
+  std::mutex                CancelDropSwitch; /* mutex for SQLCancel/SQLFreeStmt(SQL_DROP) */
+  std::atomic_bool          canceled;
   MADB_Cursor               Cursor;
-  MADB_QUERY                Query;
   MADB_List                 ListItem;
   long long                 AffectedRows= 0;
   MADB_Dbc                  *Connection;
   struct st_ma_stmt_methods *Methods;
-  Unique::ResultSet         rs;
   Unique::PreparedStatement stmt;
-  Unique::ResultSetMetaData metadata;
+  Unique::ResultSet         rs;
+  mariadb::ResultSetMetaData* metadata= nullptr;
   Unique::MYSQL_RES         DefaultsResult;
   MADB_DescRecord           *PutDataRec= nullptr;
   MADB_Stmt                 *DaeStmt= nullptr;
@@ -322,23 +325,24 @@ struct MADB_Stmt
   MADB_Desc *IArd;
   MADB_Desc *IIrd;
   MADB_Desc *IIpd;
+
   unsigned short            *UniqueIndex= nullptr; /* Insdexes of columns that make best available unique identifier */
   SQLSETPOSIROW             DaeRowNumber= 0;
   int32_t                   ArrayOffset= 0;
-  int32_t                   Status= 0;
   uint32_t                  MultiStmtNr= 0;
   uint32_t                  MultiStmtMaxParam= 0;
-  int32_t                   PutParam= -1;
+  int32_t                   PutParam= -2;
   enum MADB_StmtState       State= MADB_SS_INITED;
   enum MADB_DaeType         DataExecutionType= MADB_DAE_NORMAL;
   SQLSMALLINT               ParamCount= 0;
+
   MADB_BulkOperationInfo    Bulk;
   bool                      PositionedCommand= false;
   bool                      RebindParams= false;
   bool                      bind_done= false;
 
   MADB_Stmt(MADB_Dbc *Connection);
-  SQLRETURN Prepare(const char* StatementText, SQLINTEGER TextLength, bool ServerSide= true);
+  SQLRETURN Prepare(const char* StatementText, SQLINTEGER TextLength, bool ServerSide= true, bool DirectExecution= false);
   SQLRETURN GetOutParams(int CurrentOffset);
   SQLRETURN DoExecuteBatch();
   void AfterExecute();
@@ -399,6 +403,10 @@ void  CloseClientCharset(Client_Charset *cc);
 } while(0)
 
 #define iOdbc() (sizeof(SQLWCHAR)==4)
+#define MADB_DELETE(a) do { \
+  delete (a);\
+  (a)= nullptr; \
+} while(0)
 
 #include "ma_error.h"
 #include "ma_info.h"
