@@ -1398,6 +1398,7 @@ SQLRETURN SQL_API SQLSetParam(SQLHSTMT stmt,
 	                            SQLLEN *nval)
 {
   CHECK_STMT_CLEAR_ERROR(stmt);
+
   return MA_SQLBindParameter(stmt, par, SQL_PARAM_INPUT_OUTPUT, type, sqltype, coldef,
                              scale, val, SQL_SETPARAM_VALUE_MAX, nval);
 }
@@ -1461,6 +1462,14 @@ SQLRETURN SQL_API SQLSpecialColumns(SQLHSTMT StatementHandle,
     SQLUSMALLINT Nullable)
 {
   CHECK_STMT_CLEAR_ERROR(StatementHandle);
+#ifdef _ACTIONS_TRACE_
+#ifdef __APPLE__
+  if (getenv("GITHUB_ACTIONS") != NULL && strncmp(getenv("GITHUB_ACTIONS"), "true", 4) == 0 && IdentifierType == SQL_ROWVER)
+  {
+    printf("# -- Nullable: %hu)\n", Nullable);
+  }
+#endif // __APPLE__
+#endif
   return MA_SQLSpecialColumns(StatementHandle, IdentifierType, CatalogName, NameLength1,
     SchemaName, NameLength2, TableName, NameLength3, Scope, Nullable);
 }
@@ -1478,7 +1487,14 @@ SQLRETURN SQL_API SQLSpecialColumnsW(SQLHSTMT StatementHandle,
     SQLUSMALLINT Scope,
     SQLUSMALLINT Nullable)
 {
-  
+#ifdef _ACTIONS_TRACE_
+#ifdef __APPLE__
+  if (getenv("GITHUB_ACTIONS") != NULL && strncmp(getenv("GITHUB_ACTIONS"), "true", 4) == 0 && IdentifierType == SQL_ROWVER)
+  {
+    printf("# -- Nullable: %hu)\n", Nullable);
+  }
+#endif // __APPLE__
+#endif
   return MA_SQLSpecialColumnsW(StatementHandle, IdentifierType, CatalogName, NameLength1,
     SchemaName, NameLength2, TableName, NameLength3, Scope, Nullable);
 }
